<?php
namespace Service;

use Doctrine\ORM\EntityManager;
use Util\InstanceProvider;
use Util\Messages;
use Zeedhi\DevOps\Model\Servermanager;
use Zeedhi\DevOps\Util\Environment;
use Zeedhi\DevOps\Util\Utilities;
use Zeedhi\DevOps\Model\Configurationmanager;
use Zeedhi\DevOps\Model\Configurationmanager\Entities\ProductFiles AS ProductFilesEntity;

class ProductFiles {

    /** @var EntityManager $entityManager */
    protected $entityManager;
    /** @var Servermanager\Repositories\Server   */
    protected $serverRepository;
    /** @var Configurationmanager\Repositories\ProductFiles */
    protected $productFilesRepository;
    /** @var Configurationmanager\Repositories\Release */
    protected $releaseRepository;
    /** @var Servermanager\Repositories\ProductServer  */
    protected $productServerRepository;
    /** @var Configurationmanager\Repositories\Deploy */
    protected $deployRepository;
    /** @var Configurationmanager\Repositories\Subrelease */
    protected $subReleaseRepository;
    /** @var Environment */
    private $environment;

    /**
     * @param EntityManager $entityManager
     * @param Environment $environment
     */
    const SCRIPT_PENDING = Configurationmanager\Entities\ProductFiles::SCRIPT_PENDING;
    /**
     * @var Mnemonic
     */
    private $mnemonicService;

    public function __construct(EntityManager $entityManager, Environment $environment,Mnemonic $mnemonicService) {
        $this->entityManager = $entityManager;
        $this->environment = $environment;
        $this->mnemonicService = $mnemonicService;
        $this->serverRepository = $this->entityManager->getRepository(Utilities::SERVER_MANAGER_ENTITY_DIR . 'Server');
        $this->productFilesRepository = $this->entityManager->getRepository(Utilities::CONFIGURATION_MANAGER_ENTITY_DIR . 'ProductFiles');
        $this->releaseRepository = $this->entityManager->getRepository(Utilities::CONFIGURATION_MANAGER_ENTITY_DIR . 'Release');
        $this->subReleaseRepository = $this->entityManager->getRepository(Utilities::CONFIGURATION_MANAGER_ENTITY_DIR . 'Subrelease');
        $this->productServerRepository = $this->entityManager->getRepository(Utilities::SERVER_MANAGER_ENTITY_DIR . 'ProductServer');
        $this->deployRepository = $this->entityManager->getRepository(Utilities::CONFIGURATION_MANAGER_ENTITY_DIR . 'Deploy');
    }

    public function validationScriptApproved($productFileId, $text, $state)
    {
        if ($productFileId) {
            /** @var ProductFilesEntity $productFiles */
            $productFiles = $this->productFilesRepository->findOneBy(array('id' => $productFileId));

            if ($text != $productFiles->getText() && !($state == 'A' && $productFiles->getState() == 'P')) {
                $productFiles->setState(self::SCRIPT_PENDING);
                $this->setPendingInScripsRelated($productFiles);
                $state = self::SCRIPT_PENDING;
            }
        }
        return $state;
    }

    public function validateMnemonicWithLabel(ProductFilesEntity $productFiles){
        return $this->mnemonicService->validateMnemonicWithLabelByProductFileId($productFiles->getId());
    }

    public static function getMnemonicsByString($string)
    {
        $mnemonics = '';
        if(is_string($string)){
            $mnemonics = explode(Mnemonic::MNEMONIC_SEPARATOR, $string);
            $mnemonics = Utilities::getArrayKeyUnpaired($mnemonics,false);
        }
        return $mnemonics;
    }

    public function getAllMnemonicsByProductFileId($productFileId){
        /** @var ProductFilesEntity $productFile */
        $productFile = $this->productFilesRepository->find($productFileId);
        $this->setAllMnemonics($productFile);
    }

    public function validateMnemonicWithLabelApproved($productFileId,$state)
    {
        /** @var ProductFilesEntity $productFile */
        $productFile = $this->productFilesRepository->find($productFileId);
        $mnemonics = $this->validateMnemonicWithLabel($productFile);
        $message = null;
        if (!empty($mnemonics)) {
            $productFile->setState(ProductFilesEntity::SCRIPT_PENDING);
            $this->entityManager->persist($productFile);
            $this->entityManager->flush();
            if ($state == ProductFilesEntity::SCRIPT_APPROVED) {
                $message = 'mnemonic_not_label';
            }
        }
        return $message;
    }

    public function validateDeployServerMnemonic($productServerIds,$deployId,$typeEvent)
    {
        /** @var Servermanager\Entities\ProductServer[] $productServers */
        $productServers = $this->productServerRepository->findBy(array('id' => $productServerIds));
        /** @var Configurationmanager\Entities\Deploy $deploy */
        $deploy = $this->deployRepository->find($deployId);
        $product = $deploy->getProduct();
        $productFileTypes = Server::getScripTypesByOperation($typeEvent);
        $productFiles = $this->productFilesRepository->findBy(array('product' => $product, 'type' => $productFileTypes));
        return $this->mnemonicService->validateDeployServerMnemonic($productServers, $productFiles);
    }

    public function validateSubReleaseServerMnemonic(Configurationmanager\Entities\Subrelease $subRelease, $typeEvent)
    {
        $server = $subRelease->getServerByType($typeEvent);
        $product = $subRelease->getProduct();
        $productFileTypes = Server::getScripTypesByOperation($typeEvent);
        $productFiles = $this->productFilesRepository->findBy(array('product' => $product, 'type' => $productFileTypes));
        return $this->mnemonicService->validateServerMnemonic($server, $productFiles);
    }

    public function validateReleaseServerMnemonic(Configurationmanager\Entities\Release $release, $typeEvent)
    {
        $server = $release->getProductServer()->getServer();
        $product = $release->getProduct();
        $productFileTypes = Server::getScripTypesByOperation($typeEvent);
        $productFiles = $this->productFilesRepository->findBy(array('product' => $product, 'type' => $productFileTypes));
        return $this->mnemonicService->validateServerMnemonic($server, $productFiles);
    }

    public function findMnemonicByProductFiles($productFileId, $filter = " "){
        return $this->mnemonicService->findMnemonicByProductFiles($productFileId, $filter);
    }

    public function saveMnemonicForProduct($mnemonicId,$label, $defaultValue, $isEncrypted = 'N', $mnemonicGrouping = null){
       return $this->mnemonicService->saveMnemonicForProduct($mnemonicId, $label, $defaultValue, $isEncrypted, $mnemonicGrouping);
    }

    public function findMnemonicByProductFilesAndServer($productFileId,$serverId){
        return $this->mnemonicService->findMnemonicByProductFilesAndServer($productFileId,$serverId);
    }

    public function setMnemonicsInScriptBySubRelease($scripts,$typeEvent,Configurationmanager\Entities\Subrelease $subrelease)
    {
        $serverId = $subrelease->getProductServer()->getServer()->getId();
        $scripts = $this->replaceMnemonics($scripts, $typeEvent, $serverId);
        return $scripts;
    }

    public function setMnemonicsInScriptByReleaseId($scripts,$typeEvent, $releaseId)
    {
        /** @var Configurationmanager\Entities\Release $release */
        $release = $this->releaseRepository->find($releaseId);
        $serverId = $release->getProductServer()->getServer()->getId();
        $scripts = $this->replaceMnemonics($scripts, $typeEvent, $serverId,'ID');
        return $scripts;
    }

    public function setMnemonicsInScript($scripts, $typeEvent, $releaseId, $productServerId, $subReleaseId)
    {
        if ($productServerId) {
            /** @var Servermanager\Entities\ProductServer $productServer */
            $productServer = $this->productServerRepository->find($productServerId);
            $serverId = $productServer->getServer()->getId();
        } else if ($subReleaseId) {
            /** @var Configurationmanager\Entities\Subrelease $subRelease */
            $subRelease = $this->subReleaseRepository->find($subReleaseId);
            $serverId = $subRelease->getRelease()->getProductServer()->getServer()->getId();
        } else {
            /** @var Configurationmanager\Entities\Release $release */
            $release = $this->releaseRepository->find($releaseId);
            $serverId = $release->getProductServer()->getServer()->getId();
        }
        $scripts = $this->replaceMnemonics($scripts, $typeEvent, $serverId, 'ID');
        return $scripts;
    }

    public function setAllMnemonics(Configurationmanager\Entities\ProductFiles $productFile)
    {
        $text = $productFile->getText();
        $allMnemonicString = self::getMnemonicsByString($text);
        $this->mnemonicService->setMnemonicsProductFile($productFile, $allMnemonicString);
        return $allMnemonicString;
    }

    /**
     * @param $productFiles
     */
    protected function setPendingInScripsRelated(Configurationmanager\Entities\ProductFiles $productFiles)
    {
        /** @var Configurationmanager\Entities\ProductFiles[] $allRelatedProductFiles */
        $allRelatedProductFiles = $this->productFilesRepository
            ->findBy(array('product' => $productFiles->getProduct()));
        foreach ($allRelatedProductFiles as $relatedProductFiles) {
            $validate = strpos($relatedProductFiles->getText(), $productFiles->getFileName());
            if ($validate !== false) {
                $relatedProductFiles->setState(self::SCRIPT_PENDING);
                $this->entityManager->persist($relatedProductFiles);
            }
        }
    }

    /**
     * @param $scripts
     * @param $typeEvent
     * @param $serverId
     * @param string $indexProductFilesId
     * @return mixed
     */
    protected function replaceMnemonics($scripts, $typeEvent, $serverId,$indexProductFilesId = 'PRODUCT_FILES_ID')
    {
        $scripTypes = Server::getScripTypesByOperation($typeEvent);
        $mnemonicsUsed = array();
        foreach ($scripts as &$script) {
            if (in_array($script['TYPE'], $scripTypes)) {
                $text = $script['TEXT'];
                $mnemonics = $this->findMnemonicByProductFilesAndServer($script[$indexProductFilesId], $serverId);
                foreach ($mnemonics as $mnemonic) {
                    if($mnemonic['IS_ENCRYPTED'] == 'Y'){
                        $value = Utilities::decrypt($mnemonic['VALUE']);
                    }else {
                        $value = $mnemonic['VALUE'];
                    }
                    $value = $this->setValueMnemonics($value);
                    $text = str_replace(Mnemonic::MNEMONIC_SEPARATOR . $mnemonic['MNEMONIC'] . Mnemonic::MNEMONIC_SEPARATOR, $value, $text);
                    if(!in_array($mnemonic['MNEMONIC'], $mnemonicsUsed)) {
                        $script['MNEMONICS'][] = array('MNEMONIC' => $mnemonic['MNEMONIC'], 'VALUE' => $value);
                        $mnemonicsUsed[] = $mnemonic['MNEMONIC'];
                    }
                }
                $script['TEXT'] = $text;
            }
        }
        return $scripts;
    }

    protected function setValueMnemonics($value)
    {
        if (strtoupper($value) === 'NULL') {
            $value = '';
        } else if (strpos($value, '[NOT_ESCAPE]') === false) {
            $value = str_replace(
                array('\\', '$', '.', '*', '/', '[', ']', '^'),
                array('\\\\', '\$', '\.', '\*', '\/', '\[', '\]', '\^'),
                $value
            );
        } else {
            $value = str_replace('[NOT_ESCAPE]','',$value);
        }
        return $value;
    }

    /**
     * @param $id
     * @param $name
     * @param $productId
     * @param $fileName
     * @param $text
     * @param $relativePath
     * @param $type
     * @param $isRequired
     * @param $isWindows
     * @param $parentId
     * @param $state
     * @param $productVersionId
     * @return array
     * @throws \Doctrine\DBAL\ConnectionException
     * @throws \Exception
     */
    public function saveProductFile($id, $name, $productId, $fileName, $text, $relativePath, $type,
                                    $isRequired, $isWindows, $parentId, $state, $productVersionId) {
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try {
            $productFileClone = $productVersionClone = $textClone = null;
            /** @var ProductFilesEntity $productFile */
            $productFile = $id ? $this->productFilesRepository->find($id) : null;
            /** @var ProductFilesEntity $parent */
            $parent = $parentId ? $this->productFilesRepository->find($parentId) : null;

            if(($productFile && !$productFile->getProductVersion() && $productVersionId) ||
               ($productFile && $productVersionId && $productFile->getProductVersion() && $productFile->getProductVersion()->getId() != $productVersionId)) {
                $this->validateChildWithSameVersion($productFile, $productVersionId);
                /** @var \Zeedhi\DevOps\Model\Devops\Entities\ProductVersion $productVersionClone */
                $productVersionClone = InstanceProvider::getRepository()->getProductVersion()->find($productVersionId);
                $textClone = $text;
                $productVersionId = $productFile->getProductVersion() ? $productFile->getProductVersion()->getId() : null;
                $text = $productFile->getText();
            }

            /** @var \Zeedhi\DevOps\Model\Devops\Entities\ProductVersion $productVersion */
            $productVersion = $productVersionId ? InstanceProvider::getRepository()->getProductVersion()->find($productVersionId) : null;
            /** @var Configurationmanager\Entities\Product $product */
            $product     = InstanceProvider::getRepository()->getProduct()->find($productId);
            if($productFile){
                $oldText = $productFile->getText();  
            }else{
                $oldText = null;
            }
            $productFile = InstanceProvider::getFactory()->getProductFiles()->save($productFile, $parent, $product, $productVersion,
                $name, $fileName, $text, $relativePath, $type, $isRequired, $isWindows, $state);

            $this->entityManager->persist($productFile);
            $this->entityManager->flush();

            if($textClone) {
                $productFileClone = $this->createCloneProductFile($productFile, $productVersionClone, $textClone);
            }
            $this->updateChildProductFiles($productFile);

            $currentId = $textClone ? $productFileClone->getId() : $productFile->getId();
            $this->getAllMnemonicsByProductFileId($currentId);
            $mnemonicValidation = $this->validateMnemonicWithLabelApproved($currentId, $state);
            //$configurationLog = InstanceProvider::getFactory()->getConfigurationLog()->save($oldText, $text, 'S', $productFile, null, null);
            $date = new \DateTime();
            $mongoRow = array(
                'PRODUCT_FILE_ID'  => intval($productFile->getId()),
                'OLD_VALUE'     => $oldText,
                'NEW_VALUE' => $text,
                'NMOPERADOR'  => $this->environment->getCurrentUserName(),
                "DATE" => intval($date->format('YmdHis'))
            );
            InstanceProvider::getMongoDB()->insert("PRODUCT_FILES_LOG", $mongoRow);

            
            $connection->commit();
            return array('ID' => $currentId, 'MNEMONIC_VALIDATION' => $mnemonicValidation);
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    /**
     * @param ProductFilesEntity $productFile
     * @param $productVersionClone
     * @param $textClone
     * @return ProductFilesEntity
     */
    private function createCloneProductFile(ProductFilesEntity $productFile, $productVersionClone, $textClone) {
        $productFileClone = clone $productFile;
        $productFileClone->setId(null);
        $productFileClone->setProductVersion($productVersionClone);
        $productFileClone->setParent($productFile);
        $productFileClone->setText($textClone);
        $productFileClone->setState(self::SCRIPT_PENDING);

        $this->entityManager->persist($productFileClone);
        $this->entityManager->flush();

        return $productFileClone;
    }

    private function updateChildProductFiles(ProductFilesEntity $productFile) {
        $childrenProductFiles = $this->productFilesRepository->findBy(array('parent' => $productFile));
        /** @var ProductFilesEntity $childProductFile */
        foreach($childrenProductFiles as $childProductFile) {
            $childProductFile->setName($productFile->getName());
            $childProductFile->setFileName($productFile->getFileName());
            $childProductFile->setRelativePath($productFile->getRelativePath());
            $childProductFile->setType($productFile->getType());
            $childProductFile->setIsRequired($productFile->getIsRequired());
            $childProductFile->setIsWindows($productFile->getIsWindows());
            $this->entityManager->persist($childProductFile);
        }
        $this->entityManager->flush();
    }

    private function validateChildWithSameVersion(ProductFilesEntity $productFile, $changedProductVersionId) {
        $childrenProductFiles = $this->productFilesRepository->findBy(array('parent' => $productFile));
        /** @var ProductFilesEntity $childProductFile */
        foreach($childrenProductFiles as $childProductFile) {
            if($childProductFile->getProductVersion() && $childProductFile->getProductVersion()->getId() == $changedProductVersionId) {
                throw new \Exception(Messages::ERR_EXISTS_CHILD_WITH_VERSION);
            }
        }
    }

    /**
     * @param $productId
     * @param $productVersionId
     * @param $allowEmptyFilter
     * @return array
     */
    public function findProductFilesToExecuteByVersion($productId, $productVersionId, $allowEmptyFilter) {
        if(empty($productVersionId) && !$allowEmptyFilter) {
            return array();
        }
        $version = $revision = null;
        if($productVersionId) {
            /** @var \Zeedhi\DevOps\Model\Devops\Entities\ProductVersion $productVersion */
            $productVersion = InstanceProvider::getRepository()->getProductVersion()->find($productVersionId);
            $version  = $productVersion->getVersion();
            $revision = $productVersion->getRevision();
        }
        return $this->productFilesRepository->findProductFilesToExecuteByVersion($productId, $version, $revision);
    }

    /**
     * @param $productId
     * @param $productVersionId
     * @param $typeEvent
     * @param $productServerId
     * @param $conditions
     * @return array
     */
    public function findVersionProductFiles($productId, $productVersionId, $typeEvent, $productServerId, $conditions) {
        $version = $revision = null;
        if($productVersionId) {
            /** @var \Zeedhi\DevOps\Model\Devops\Entities\ProductVersion $productVersion */
            $productVersion = InstanceProvider::getRepository()->getProductVersion()->find($productVersionId);
            $version  = $productVersion->getVersion();
            $revision = $productVersion->getRevision();
        }
        $filter = Utilities::createQueryFiltered($conditions);
        $scripts = $this->productFilesRepository->findVersionProductFiles($productId, $version, $revision, $filter);
        if ($typeEvent) {
            $scripts = $this->setMnemonicsInScript($scripts, $typeEvent, null, $productServerId, null);
        }
        return $scripts;
    }
}
